/*
 *  Author: Erik Ruggles
 *  compilation: gcc -o nfc rbm.c netflixcore.c
 *  Purpose:  To determine unknown movie ratings for users using the Netflix dataset.
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "netflixcore.h"
#define SOFTMAX		5
#define FEATURES	100

//Constants as described in Hinton 2007

#define BATCH_SIZE	1000
#define lrw			0.01
#define lrv			0.01
#define lrh			0.01
#define momentum	0.9
#define weightcost	0.001


//Constants as described in http://code.google.com/p/nprizeadditions/
/*
#define BATCH_SIZE	100
#define lrw			0.001
#define lrv			0.008
#define lrh			0.0006
#define momentum	0.8
#define weightcost	0.0001
*/

//Other constants
/*
#define BATCH_SIZE	1000
#define lrw			1.0
#define lrv			1.0
#define lrh			1.0
#define momentum	0.9
#define weightcost	0.001
*/

#define numLoops	2
#define ZERO(v)		memset(v,0,sizeof(v))

//The RBM and its counters
double VisBiases[NMOVIES][SOFTMAX];
double HidBiases[FEATURES];
double Weights[NMOVIES][SOFTMAX][FEATURES];

int negVisStates[NMOVIES];

int posHidStates[FEATURES];
int negHidStates[FEATURES];

double posVis[NMOVIES][SOFTMAX];
double negVis[NMOVIES][SOFTMAX];
double incVis[NMOVIES][SOFTMAX];

double posHid[FEATURES];
double negHid[FEATURES];
double incHid[FEATURES];

double cdPos[NMOVIES][SOFTMAX][FEATURES];
double cdNeg[NMOVIES][SOFTMAX][FEATURES];
double incCD[NMOVIES][SOFTMAX][FEATURES];

double posHidProbs[FEATURES];
double negHidProbs[FEATURES];
double negVisProbs[NMOVIES][SOFTMAX];
double negVisProbsRMSE[NMOVIES][SOFTMAX];

int cultMovies[31];

int popCult() {
	//Hardcoded list of cult movies
	cultMovies[0] = 9340;
	cultMovies[1] = 5317;
	cultMovies[2] = 6386;
	cultMovies[3] = 14454;
	cultMovies[4] = 7635;
	cultMovies[5] = 4266;
	cultMovies[6] = 8367;
	cultMovies[7] = 2939;
	cultMovies[8] = 14042;
	cultMovies[9] = 11022;
	cultMovies[10] = 14916;
	cultMovies[11] = 6174;
	cultMovies[12] = 6296;
	cultMovies[13] = 16734;
	cultMovies[14] = 12130;
	cultMovies[15] = 3151;
	cultMovies[16] = 2189;
	cultMovies[17] = 8;
	cultMovies[18] = 8131;
	cultMovies[19] = 12660;
	cultMovies[20] = 7904;
	cultMovies[21] = 3565;
	cultMovies[22] = 12232;
	cultMovies[23] = 1607;
	cultMovies[24] = 15482;
	cultMovies[25] = 14890;
	cultMovies[26] = 14274;
	cultMovies[27] = 9197;
	cultMovies[28] = 11819;
	cultMovies[29] = 8782;
	cultMovies[30] = 17062;

	int i = 0;
	for (i = 0; i < 31; i++)
		cultMovies[i] -= 1;

	return 1;
}

int main(int argc, char** argv) {
	//Load the Database
	nf_allocate();
	nf_load();
	nf_hideprobeset();
	
	//Setup and run the RBM
	popCult();
	initWeights();
	initBiases();
	train();
	dump();
}

double randn() {
	//Generate random number between 0 and 1
	return (rand()/(double)(RAND_MAX));
}


int train() {
	/*************
	* Runs the RBM.
	*************/
	
	//Setup constants and counters
	double rmse = 0.0;
	double crmse = 0.0;
	double srmse = 0.0;
	double prmse = 0.0;
	long int u = 0;
	int c,i,j,k,m,n,r = 0;
	double epsW = lrw;
	double epsV = lrv;
	double epsH = lrh;
	double Momentum = momentum;
	double WeightCost = weightcost;
	int movieC[NMOVIES];
	time_t start;
	struct tm *timeinfo;

	
	//Usefull for when reviewing outputs
	printf("The following are the constants used:\n");
	printf("lrw = %f\n",lrw);
	printf("lrv = %f\n",lrv);
	printf("lrh = %f\n",lrh);
	printf("wc = %f\n",weightcost);
	printf("M = %f\n",momentum);
	printf("loops = %d\n",numLoops);
	printf("feat = %d\n",FEATURES);
	printf("bSize = %d\n",BATCH_SIZE);

	//Clear matrices
	ZERO(incHid);
	ZERO(incVis);
	ZERO(incCD);
	ZERO(movieC);
	
	//Begin the training!
	for (c = 0; c < numLoops; c++) {
	
		//Reset/Update variables/matrices
		if(c > 6)
			Momentum = .9;
		time(&start);
		timeinfo = localtime(&start);
		printf("%s",asctime(timeinfo));
		int numRated = 0;
		int cnumRated = 0;
		int snumRated = 0;
		int pnumRated = 0;
		rmse = 0.0;
		crmse = 0.0;
		srmse = 0.0;
		prmse = 0.0;
		ZERO(cdNeg);
		ZERO(cdPos);
		ZERO(posHid);
		ZERO(negHid);
		ZERO(posVis);
		ZERO(negVis);
		ZERO(movieC);
		
		
		for (u = 0; u < NUSERS; u++) {
			double sumWeights[FEATURES];
			ZERO(negVisProbsRMSE);
			ZERO(negVisProbs);
			ZERO(sumWeights);
			rlist *userRatings = nf_ratingsforuser(u);
			int rated = userRatings->length;
			
			//Used to run on low rating users only
			//if (rated >= 20)
			//	continue;
			
			//Set visible side and start accumulating for vis bias update and 
			//hidden node calculation
			for (i = 0; i < rated; i ++) {
				if (userRatings->ratings[i].visible == 1) {
					m = userRatings->ratings[i].id;
					r = userRatings->ratings[i].stars - 1;
					posVis[m][r] += 1.0;
					movieC[m] += 1;
					numRated += 1;
					
					for (j = 0; j < FEATURES; j ++) {
						sumWeights[j] += Weights[m][r][j];
					}
				}
			}
		
			//P(h = 1 | V)
			for (j = 0; j < FEATURES; j++) {
				posHidProbs[j] = 1.0/(1.0 + exp(-sumWeights[j] - HidBiases[j]));
				
				if (posHidProbs[j] > randn()) {
					posHidStates[j] = 1;
					posHid[j] += 1.0;
				} else {
					posHidStates[j] = 0;
				}
			}
			
			//Pass the current hidden states along to the hidden side as we want to 
			//keep the current pos states for later
			for (j = 0; j < FEATURES; j++) {
				negHidStates[j] = posHidStates[j];
			}
			
			//Determine the size of T
			int tSteps = 0;
			if (c < 18)
				tSteps = 1;
			else if (c < 27)
				tSteps = 3;
			else if (c < 40)
				tSteps = 5;
			else
				tSteps = 9;
			//int tSteps = 1 + (c / 2);
			int t = 0;

			//Start the Negative side
			for (t = 0; t < tSteps; t++) {
				//Reset matrices
				ZERO(negVisStates);	
				ZERO(sumWeights);
				ZERO(negVisProbsRMSE);
				
				//Calculate Visual Side of Graph
				for (i = 0; i < rated; i++) {
					m = userRatings->ratings[i].id;		
					for (j = 0; j < FEATURES; j++) {
						if (negHidStates[j] == 1) {
							for (k = 0; k < SOFTMAX; k++) {
								negVisProbs[m][k] += Weights[m][k][j];
							}
						}				
						
						if (t == 0) {
							for (k = 0; k < SOFTMAX; k++) {
								negVisProbsRMSE[m][k] += posHidProbs[j] * Weights[m][k][j];
							}
						}
					}

					if (t == 0) {
						double totalProbsRMSE = 0.0;
						for (k = 0; k < SOFTMAX; k++) {
							negVisProbsRMSE[m][k] = 1.0/(1 + exp(-negVisProbsRMSE[m][k] - VisBiases[m][k]));
							totalProbsRMSE += negVisProbsRMSE[m][k];
						}
						if (totalProbsRMSE != 0.0) {
							for(k = 0; k < SOFTMAX; k++)
								negVisProbsRMSE[m][k] /= totalProbsRMSE;
						}
					}
					
					if (userRatings->ratings[i].visible == 1) {
					
						//P(Vk = 1 | H)
						double totalProbs = 0;
						for (k = 0; k < SOFTMAX; k++)	{
							negVisProbs[m][k] = 1.0/(1 + exp(-negVisProbs[m][k] - VisBiases[m][k]));
							totalProbs += negVisProbs[m][k];
						}
						if (totalProbs != 0.0) {
							for (k = 0; k < SOFTMAX; k++) {
								negVisProbs[m][k] /= totalProbs;
							}
						}
						//Determine visible side
						double randomNum = randn();
						if ((randomNum -= negVisProbs[m][0]) <= 0.0)
							negVisStates[m] = 0;
						else if ((randomNum -= negVisProbs[m][1]) <= 0.0)
							negVisStates[m] = 1;
						else if ((randomNum -= negVisProbs[m][2]) <= 0.0)
							negVisStates[m] = 2;
						else if ((randomNum -= negVisProbs[m][3]) <= 0.0)
							negVisStates[m] = 3;
						else
							negVisStates[m] = 4;
							
						//Since this is the last step, accumulate negative for vis bias update
						if (t + 1 >= tSteps)
							negVis[m][negVisStates[m]] += 1.0;
							
						for (j = 0; j < FEATURES; j++) {
							sumWeights[j] += Weights[m][negVisStates[m]][j];
						}
					}
				}
				
				//Calc RMSE				
				if (t == 0) {
					for(i = 0; i < rated; i++) {
						//Find the error
						m = userRatings->ratings[i].id;
						r = userRatings->ratings[i].stars-1;
						double expV = 0.0;
						for (k = 0; k < SOFTMAX; k++) {
							expV += (((double)k) * negVisProbsRMSE[m][k]);
						}
						double diff = ((double)r) - expV;
						
						if (userRatings->ratings[i].visible == 0) {
						//Sum up low ratings users
						if (rated < 20) {
							srmse += (diff*diff);
							snumRated += 1;
						}
						
						//Sum up cult movies
						int q = 0;
						for (q = 0; q < 31; q++) {
							if (cultMovies[q] == m) {
								crmse += (diff * diff);
								cnumRated += 1;
							}
						}
						
						//Sum up probe ratings
						//if (userRatings->ratings[i].visible == 0) {
							prmse += (diff*diff);
							pnumRated += 1;
						} else {
						//Sum up overall ratings
							rmse += (diff*diff);
						}
					}
				}
				
				//Calc Hid Layer and if last step accumulate negative info for hid bias update
				for (j = 0; j < FEATURES; j++) {
					negHidProbs[j] = 1.0/(1.0 + exp(-sumWeights[j] - HidBiases[j]));
					if (negHidProbs[j] > randn()) {
						negHidStates[j] = 1;
						if (t + 1 >= tSteps)
							negHid[j] += 1.0;
					} else {
						negHidStates[j] = 0;
					}
				}

				ZERO(negVisProbs);
			}
			
			//Accumulate both positive and negative sides of CD
			for (i = 0; i < rated; i++) {
				if (userRatings->ratings[i].visible) {
					m = userRatings->ratings[i].id;
					r = userRatings->ratings[i].stars - 1;
					for (j = 0; j < FEATURES; j++) {
						cdPos[m][r][j] += (double)posHidStates[j];
						cdNeg[m][negVisStates[m]][j] += (double)negHidStates[j];
					}
				}
			}
			
			//Update after each batch
			if (((u + 1) % BATCH_SIZE) == 0 || (u+1) == NUSERS) {
				int bSize = (u % BATCH_SIZE) + 1;

				// dH = <H_j>_0 - <H_j>_T
				for (j = 0; j < FEATURES; j++) {
					if (posHid[j] != 0.0 || negHid[j] != 0.0) {
						incHid[j] = Momentum * incHid[j] + epsH * ((posHid[j] - negHid[j]) / (double) bSize);
						HidBiases[j] += incHid[j];
					}
				}
				for (m = 0; m < NMOVIES; m++) {
					if (movieC[m] == 0)
							continue;
					for (k = 0; k < SOFTMAX; k++) {
						// dV = <V^m_k>_0 - <V^m_k>_T
						if (negVis[m][k] != 0.0 || posVis[m][k] != 0.0) {
							incVis[m][k] = Momentum * incVis[m][k] + epsV * ((posVis[m][k] - negVis[m][k]) / (double)movieC[m]);
							VisBiases[m][k] += incVis[m][k];
						}
						// dW = <V^m_k H_j>_0 - <V^m_k H_j>_T
						for (j = 0; j < FEATURES; j++) {
							if (cdPos[m][k][j] != 0.0 || cdNeg[m][k][j] != 0.0) {
								incCD[m][k][j] = Momentum * incCD[m][k][j] + epsW * (((cdPos[m][k][j] - cdNeg[m][k][j]) / (double)movieC[m]) - WeightCost * Weights[m][k][j]);
								Weights[m][k][j] += incCD[m][k][j];
							}
						}
					}
				}

				ZERO(negHid);
				ZERO(posHid);
				ZERO(negVis);
				ZERO(posVis);
				ZERO(cdPos);
				ZERO(cdNeg);
				ZERO(movieC);
			}
		}

		rmse = sqrt(rmse/(double)numRated);
		crmse = sqrt(crmse/(double)cnumRated);
		srmse = sqrt(srmse/(double)snumRated);
		prmse = sqrt(prmse/(double)pnumRated);
		printf("The RMSE for iteration %d is %f with Cult movies at %f, low ratings at %f, and a probe RMSE of %f\n",c,rmse,crmse,srmse,prmse);
	}
}

int arrayLengths() {
	//Used to count up number of users with less than 50 rated movies
	//No Longer in Use
	int counts[50];
	int i = 0;
	int total = 0;
	ZERO(counts);
	for (i = 0; i < NUSERS; i++) {
		rlist *userRatings = nf_ratingsforuser(i);
		if (userRatings->length < 50) {
			counts[userRatings->length] += 1;
			total += 1;
		}
	}
	for (i = 0; i < 50; i++) {
		printf("%d users have rated %d movies\n",counts[i],i);
	}
	printf("%f users have rated under 50 movies\n",((double)total/(double)NUSERS));
	return 1;
}

int initWeights() {
	//Initialize weights as described in Hinton 2007
	int m,j,k = 0;
	for (m = 0; m < NMOVIES; m++) {
		for (j = 0; j < FEATURES; j++) {
			for (k = 0; k < SOFTMAX; k++)
				Weights[m][k][j] = 0.02 * randn() - 0.01;
		}
	}

	return 1;
}

int initBiases() {
	//Initialize biases as described in Hinton 2007
	int m,i,k,j = 0;
	double movieCounts[NMOVIES][SOFTMAX];
	double movieRatings[NMOVIES];
	ZERO(HidBiases);
	ZERO(VisBiases);
	ZERO(movieCounts);
	ZERO(movieRatings);
	printf("Starting to count stuff\n");
	for (m = 0; m < NMOVIES; m++) {
		rlist *userRatings = nf_ratingsformovie(m);
		for (i = 0; i < userRatings->length; i++) {	
			movieCounts[m][userRatings->ratings[i].stars - 1] += 1;
			movieRatings[m] += 1;
		}
	}
	for (m = 0; m < NMOVIES; m++) {
		for (k = 0; k < SOFTMAX; k++) {
			VisBiases[m][k] = log( (double)movieCounts[m][k] / (double)movieRatings[m]);
		}
	}
	for (j = 0; j < FEATURES; j++)
		HidBiases[j] = 0.0;
	return 1;
}

int dump() {
	int i,j,k;
	FILE *file;
	file = fopen("rbm_values.bin","w");
	fwrite(Weights, sizeof(Weights[0][0][0]), sizeof(Weights)/sizeof(Weights[0][0][0]), file);
	fwrite(Weights, sizeof(HidBiases[0]), sizeof(HidBiases)/sizeof(HidBiases[0]), file);
	fwrite(Weights, sizeof(VisBiases[0][0]), sizeof(VisBiases)/sizeof(VisBiases[0][0]), file);
}
